@@ -19,7 +19,7 @@
 #ifndef _LVM_LIB_H
 #define _LVM_LIB_H
 
-#include <configure.h>
+#include "configure.h"
 
 #define _REENTRANT
 #define _GNU_SOURCE
@@ -28,6 +28,7 @@
 #include "log.h"
 #include "intl.h"
 #include "lvm-types.h"
+#include "lvm-wrappers.h"
 
 #include <libdevmapper.h>
 
